#
# act.py - DITrack 'act' command
#
# Copyright (c) 2006 The DITrack Project, www.ditrack.org.
#
# $Id: act.py 602 2006-09-29 18:27:02Z oleg $
# $HeadURL: file:///X/svn/ditrack/src/tags/0.4/DITrack/Command/act.py $
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, 
# this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice, 
# this list of conditions and the following disclaimer in the documentation 
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#

import copy
import email
import os
import sys

# DITrack modules
import DITrack.Command.generic
import DITrack.DB
import DITrack.Edit
import DITrack.UI
import DITrack.Util.common

class Handler(DITrack.Command.generic.Handler):
    canonical_name = "act"
    description = """Perform actions on an issue (or multiple issues).
usage: %s ISSUENUM [ISSUENUM...]""" % canonical_name

    def run(self, opts, globals):
	self.check_options(opts)

	if len(opts.fixed) < 2:
	    self.print_help(globals)
	    sys.exit(1)

	# We'll need an editor.
	globals.get_editor()

	db = DITrack.Util.common.open_db(opts)

	issues_list = opts.fixed[1:]

	issue = []
	issue_prev = []
	present_categories = {}
	present_vsets = {}

	for idx, i in enumerate(issues_list):
	    # Check if the issue numbers are correct.
	    try:
		issue.append(DITrack.DB.Issue.fromfile(globals, db, i))
	    except ValueError:
		DITrack.Util.common.err("Invalid issue number '" + i + "'")

	    present_categories[issue[idx].info["Category"]] = 1

	    present_vsets[			\
		    db.category[		\
		    issue[idx].info["Category"]	\
		    ].version_set] = 1

	    # Save previous issue state, to be able to identify changes.
	    issue_prev.append(copy.deepcopy(issue[idx]))

	same_category = (len(present_categories.keys()) == 1)
	same_vset = (len(present_vsets.keys()) == 1)
	single_issue = (len(issue) == 1)

	# Build up the menu.
	mi_abort = DITrack.UI.MenuItem("a", "abort, discarding changes")
	mi_ch_due_in = DITrack.UI.MenuItem("d", "change due version")
	mi_close = DITrack.UI.MenuItem("c", "close the issue")
	mi_edit_info = DITrack.UI.MenuItem("h", "edit the issue header")
	mi_edit_text = DITrack.UI.MenuItem("e", "edit comment text")
	mi_quit = DITrack.UI.MenuItem("q", "quit, saving changes")
	mi_reassign = DITrack.UI.MenuItem("o", "reassign the issue owner")
	mi_reopen = DITrack.UI.MenuItem("r", "reopen the issue")

	menu = DITrack.UI.Menu("Choose an action for the issue(s)", \
		[		\
		mi_abort, 	\
		mi_ch_due_in, 	\
		mi_close, 	\
		mi_edit_info, 	\
		mi_edit_text, 	\
		mi_quit,	\
		mi_reassign, 	\
		mi_reopen	\
		])

	save_changes = 0

	comment_text = ""

	mi_ch_due_in.enabled = same_vset
	mi_edit_info.enabled = single_issue

	while 1:

	    # Conditionally enable/disable menu items.
	    mi_close.enabled = filter(lambda x: x.info["Status"] == "open", 
		    issue)

	    mi_reopen.enabled = filter(lambda x: x.info["Status"] == "closed", 
		    issue)

	    r = menu.run()
	    if r == mi_abort:
		break

	    elif r == mi_close:
		for i in issue:
		    i.set_info_header("Status", "closed")

	    elif r == mi_ch_due_in:
		assert(same_vset)

		if single_issue:
		    print "Current due version: " + issue[0].info["Due-in"]

		due_menu = DITrack.UI.EnumMenu( 	\
			"Choose new due version",	\
			issue[0].versions["future"])

		v = due_menu.run()

		if not v: break

		for i in issue:
		    i.set_info_header("Due-in", v)

	    elif r == mi_edit_info:
		info = copy.deepcopy(issue[0].info)

		# Only header, not the description.
		info.set_payload("")

		header = \
			DITrack.Edit.edit_text(globals,	\
			info.as_string())

		descr = issue[0].info.get_payload()

		issue[0].info = email.message_from_string(header)
		issue[0].info.set_payload(descr)

	    elif r == mi_edit_text:
		comment_text = DITrack.Edit.edit_text(globals,
				comment_text)

	    elif r == mi_quit:
		save_changes = 1
		break

	    elif r == mi_reassign:

		if single_issue:
		    print "Current issue owner: " + issue[0].info["Owned-by"]

		users = db.users.keys()
		users.sort()
		owner_menu = DITrack.UI.EnumMenu( 	\
			"Choose new issue owner",	\
			users)

		v = owner_menu.run()

		if not v: break

		for i in issue:
		    i.set_info_header("Owned-by", v)

	    elif r == mi_reopen:
		for i in issue:
		    i.set_info_header("Status", "open")

	    else:
		raise NotImplementedError

	if save_changes:
	    ops = []

	    log_msgs = {}
	    comment_num = {}

	    for idx in range(len(issue)):
		comment = DITrack.DB.Comment(db, issue_prev[idx], issue[idx], \
		    comment_text, globals)

		changed_headers = comment.get_changed_headers()

		ops.extend(comment.write())

		log_actions = []

		if changed_headers.has_key("Status"):

		    assert(issue[idx].info.has_key("Status"))

		    if issue[idx].info["Status"] == "open":
			log_actions.append("reopened")

		    elif issue[idx].info["Status"] == "closed":
			log_actions.append("closed")

		    else:
			raise NotImplementedError("unknown status: " + \
			    issue[idx].info["Status"])


		if changed_headers.has_key("Owned-by"):

		    assert(issue[idx].info.has_key("Owned-by"))

		    log_actions.append("reassigned to " \
			+ issue[idx].info["Owned-by"])


		if changed_headers.has_key("Due-in"):

		    assert(issue[idx].info.has_key("Due-in"))

		    log_actions.append("moved to " + issue[idx].info["Due-in"])

		if changed_headers.has_key("Title"):

		    assert(issue[idx].info.has_key("Title"))

		    log_actions.append("title changed")


		# changed_headers gets desctructed here!

		changed_headers = filter( \
		    lambda x:
			(x != "Status") and (x != "Owned-by") and   \
				(x != "Due-in") and (x != "Title")  \
		    ,
		    changed_headers.keys())

		if changed_headers:
		    log_actions.append("header changed")

		if comment_text != "":
		    log_actions.append("comment added")

		if len(log_actions):
		    log_msgs[issue[idx].id] = ", ".join(log_actions)
		    comment_num[issue[idx].id] = issue[idx].comments_qty + 1

	    # Now create fancy log message.
	    msg = []
	    ids = log_msgs.keys()
	    ids.sort()

	    while ids:
		title = "i#%d" % ids[0]

		for seq in range(1, len(ids)):

		    if ids[seq] == -1:
			continue

		    if log_msgs[ids[seq]] == log_msgs[ids[0]]:

			title = title + (", i#%d" % ids[seq])

			ids[seq] = -1

		msg.append(title + ": " + log_msgs[ids[0]])

		ids[0] = -1
		ids = filter(lambda x: x != -1, ids)

	    DITrack.Util.common.svn_commit(globals, opts, ops, "\n".join(msg))

	    lst = comment_num.keys()  
	    lst.sort()
	    for issue in lst:
		print "Comment %d added to issue %d" % \
		    (comment_num[issue], issue)
