#
# list.py - DITrack 'list' command
#
# Copyright (c) 2006 The DITrack Project, www.ditrack.org.
#
# $Id: list.py 550 2006-09-26 00:05:45Z vss $
# $HeadURL: file:///X/svn/ditrack/src/tags/0.4/DITrack/Command/list.py $
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, 
# this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice, 
# this list of conditions and the following disclaimer in the documentation 
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#

import os
import re

# DITrack modules
import DITrack.Command.generic

class FilterExpressionError(Exception):
	"Syntax error in filter expression"
	pass

class FilterUnknownError(Exception):
	"Unknown predefined filter name"
	pass

class Filter:
	def __init__(self, db, str):
		"Initialize filter by parsing the expression passed"

		# Is it a predefined filter name?
		inplace_re = re.compile("[,=]")

		if not inplace_re.search(str):
			# Looks like a filter name
			if not db.filter.has_key(str):
				raise FilterUnknownError(str)

			str = db.filter[str]

		# Split into subclauses.
		clauses = filter(lambda x: len(x), str.split(","))

		self.conditions = []
		condition_re = re.compile("^(.*[^!=])(!?=)(.*)$")

		for c in clauses:
			m = condition_re.match(c)
			if not m:
				raise FilterExpressionError(c)

			# Perform substitutions.
			value = self.substitute(m.group(3))

			self.conditions.append((m.group(1), m.group(2), \
				value))

	def substitute(self, str):
		
		out = ""
		idx = 0
		l = len(str)
		while idx < l:
			dollar = str.find("$", idx)
			if dollar < 0:
				out = out + str[idx:]
				break
			elif dollar == (l - 1):
				break
			elif str[dollar + 1] == "$":
				out = out + "$"
				idx = dollar + 2
			else:
				end = dollar + 1
				while (end < l) and 		\
					(str[end].isalnum() or 	\
						str[end] == "_"):
					end = end + 1

				if end != (dollar + 1):
					var = str[dollar + 1:end]
					if os.environ.has_key(var):
						out = out + os.environ[var]

				idx = end

		return out


	def matches(self, issue):
		for c in self.conditions:
			assert(len(c) == 3)

			if c[1] == "=":
				if issue.info[c[0]] != c[2]: return None
			elif c[1] == "!=":
				if issue.info[c[0]] == c[2]: return None
			else:
				raise NotImplemented(c[1])

		return True

class Handler(DITrack.Command.generic.Handler):
	canonical_name = "list"
	description = """List issues, possibly filtering.
usage: %s [FILTER...]""" % canonical_name

	def run(self, opts, globals):
		self.check_options(opts)

		db = DITrack.Util.common.open_db(opts)

		filters = []
		for e in opts.fixed[1:]:
			try:
				f = Filter(db, e)
			except DITrack.Command.list.FilterUnknownError:
				DITrack.Util.common.err("ERROR: filter '%s' " \
					"is not defined" % e)

			filters.append(f)

		for i in range(1, db.next_issue_id):
			try:
				issue = DITrack.DB.Issue.fromfile(globals, \
					db, i)
			except IOError:
				DITrack.Util.common.err( \
					"ERROR: cannot open issue %d" % i, \
					fatal=None)
				continue
	
			matches = None
			for f in filters:
				if f.matches(issue):
					matches = True

			if len(filters) and not matches: continue

			print "%-4d %-8s %-8s %-6s %s" % (	\
				i,				\
				issue.info["Owned-by"],		\
				issue.info["Due-in"],		\
				issue.info["Status"],		\
				issue.info["Title"]		\
				)
