#
# DB.py - database interface
#
# Copyright (c) 2006 The DITrack Project, www.ditrack.org.
#
# $Id: DB.py 553 2006-09-28 03:15:37Z vss $
# $HeadURL: file:///X/svn/ditrack/src/tags/0.4/DITrack/DB.py $
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, 
# this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice, 
# this list of conditions and the following disclaimer in the documentation 
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#

import email
import email.Message
import re
import os.path

# DITrack modules
import DITrack.SVN

# Root database directory property containing format version.
VERSION_PROP = "ditrack:format"

# Current database format version.
FORMAT_VERSION = "1"

class NotDirectoryError(Exception):
	message = "Database path is not an [existing] directory"

class NotDatabaseError(Exception):
	message = "Path specified is not an issue database root"

class InvalidVersionError(Exception):
	message = "Database format is not supported"

class CorruptedDBError(Exception):
	message = "Database is corrupted"

class IssueExistsError(Exception):
	message = "Cannot add new issue: correponding directory already exists"

class Versions:
	def __init__(self, path):
		f = open(path)
		versions = email.message_from_file(f)
		f.close()

		self.set = {}
		for s in versions.keys():
			v = versions[s].strip().split("/")
			if len(v) != 3:
				raise CorruptedDBError("Invalid version " + \
					"set '" + s + "' definition")

			name = [ "past", "current", "future" ]
			self.set[s] = {}
			for i in range(3):
				list = v[i].strip().split()
				self.set[s][name[i]] = list

class Category:
	def __init__(self, version_set, versions, default_owner):
		self.version_set = version_set
		self.versions = versions
		self.default_owner = default_owner

class db:
	def __init__(self, path):
		# Check that it's a directory.
		if not os.path.isdir(path):
			raise NotDirectoryError

		# Get database version
		try:
			v = DITrack.SVN.propget(VERSION_PROP, path)
		except:
			raise

		if not v or not len(v):
			raise NotDatabaseError

		# Check if this is supported version.
		if v != FORMAT_VERSION:
			raise InvalidVersionError

		# Prepare pathnames array.
		self.path = {}
		self.path["/"] = path
		self.path["categories"] = \
			os.path.join(path, "etc", "categories")
		self.path["data"] = os.path.join(path, "data")
		self.path["filters"] = \
			os.path.join(path, "etc", "filters")
		self.path["next-issue-id"] = \
			os.path.join(path, "meta", "next-id")
		self.path["users"] = \
			os.path.join(path, "etc", "users")
		self.path["versions"] = \
			os.path.join(path, "etc", "versions")

		# Figure out next available issue id.
		if not os.path.exists(self.path["next-issue-id"]):
			raise CorruptedDBError(self.path["next-issue-id"] + \
				" doesn't exist")

		f = open(self.path["next-issue-id"])
		s = f.readline()
		if not s:
			f.close()
			raise CorruptedDBError(self.path["next-issue-id"] + \
				" is empty")

		try:
			self.next_issue_id = int(s.strip())
		except ValueError:
			raise CorruptedDBError("Contents of " + \
				self.path["next-issue-id"] + " are invalid")

		# Parse users.
		self.users = self.parse_users(self.path["users"])

		# Parse versions.
		self.versions = DITrack.DB.Versions(self.path["versions"])

		# Parse categories.
		f = open(self.path["categories"])
		categories = email.message_from_file(f)
		f.close()

		self.category = {}
		for c in categories.keys():
			s = categories[c].strip().split()
			if len(s) != 2:
				raise CorruptedDBError("Invalid category '" + \
					c + "' definition")

			attr = {}
			for p in s:
				av = p.split("=")
				if len(av) != 2:
					raise CorruptedDBError("Invalid " + \
						"category '" + c + \
						"' definition (attribute/" \
						"value pair)")

				attr[av[0]] = av[1]

			if not attr.has_key("versions"):
				raise CorruptedDBError("Invalid category '" + \
					c + "' definition: no versions" + \
					" defined")

			if not self.versions.set.has_key(attr["versions"]):
				raise CorruptedDBError("Invalid category '" + \
					c + "' definition: version set '" + \
					attr["versions"] + "' undefined")

			if not attr.has_key("default-owner"):
				raise CorruptedDBError("Invalid category '" + \
					c + "' definition: no default owner" \
					+ " defined")

			if not self.users.has_key(attr["default-owner"]):
				raise CorruptedDBError("Invalid category '" + \
					c + "' definition: default owner " + \
					"'" + attr["default-owner"] + \
					"' undefined")

			self.category[c] = Category( \
				attr["versions"],			\
				self.versions.set[attr["versions"]],	\
				attr["default-owner"] \
				)


		if not os.path.exists(self.path["filters"]):
			# No predefined filters.
			self.filter = {}
		else:
			f = open(self.path["filters"])
			self.filter = email.message_from_file(f)
			f.close()

	def add_issue(self):
		"Add new issue to the database"
	
		path = os.path.join(self.path["data"], \
			"i%d" % self.next_issue_id)

		if os.path.exists(path):
			raise IssueExistsError(self.next_issue_id)

		os.mkdir(path)

		# Update next-issue-id.
		self.next_issue_id = self.next_issue_id + 1

		f = open(self.path["next-issue-id"], "w")
		f.write("%d\n" % self.next_issue_id)
		f.close()

		return (self.next_issue_id - 1,	\
			[ \
			DITrack.SVN.Operation (DITrack.SVN.OP_ADD, path),
			DITrack.SVN.Operation (DITrack.SVN.OP_CHANGE, \
				self.path["next-issue-id"])	\
			])


	def is_valid_issue_num(self, issue_num):
		return (issue_num > 0) and (issue_num < self.next_issue_id)

	def parse_users(self, fname):
		users = {}
		f = open(fname)
		while 1:
			str = f.readline()
			if not str: break

			user = str.strip()
			users[user] = user

		f.close()

		return users

class Issue:
	def __init__(self, globals, db):
		self.db = db
		self.globals = globals

		self.svn_ops = []
		self.fname = {}
		self.comment = []

		# Was the issue created anew?
		self.created = None

	def fromfile(cls, globals, db, issue_id):
		"Create the issue instance reading data from a disk"

		issue = cls(globals, db)

		id = int(issue_id)
		issue.set_id(id)

		if not db.is_valid_issue_num(id):
			raise ValueError

		# Read in the issue header and description.
		f = open(issue.fname["info"])
		issue.info = email.message_from_file(f)
		f.close()

		issue.versions = db.category[issue.info["Category"]].versions

		# Count number of comments present in the issue.
		comment_re = re.compile("^comment(\\d+)$")
		issue.comments_qty = 0
		for fn in os.listdir(issue.path):
			m = comment_re.match(fn)
			if m:
				n = int(m.group(1))
				if n > issue.comments_qty:
					issue.comments_qty = n
	
		for cid in range(1, issue.comments_qty + 1):
			fname = "%s%d" % (issue.fname["comment"], cid)
			
			f = open(fname)
			issue.comment.append(email.message_from_file(f))
			f.close()

		return issue

	fromfile = classmethod(fromfile)

	def create(cls, globals, db, title, opened_by, category, \
		version_reported, version_due, description):
		"Create new issue from scratch"

		assert(db.category.has_key(category))

		issue = cls(globals, db)

		issue.info = email.Message.Message()

		issue.info["Opened-by"] = opened_by
		issue.info["Opened-on"] = \
			DITrack.Util.common.fmt_timestamp(issue.globals)
		issue.info["Owned-by"] = db.category[category].default_owner
		issue.info["Title"] = title
		issue.info["Category"] = category
		issue.info["Status"] = "open"
		issue.info["Reported-for"] = version_reported
		issue.info["Due-in"] = version_due

		issue.info.set_payload(description)

		issue.created = True
		return issue

	create = classmethod(create)

	def add_comment(self, comment_text):
		"Add a comment to the issue."
		fname = "%s%d" % (self.fname["comment"], self.comments_qty + 1)
		f = open(fname, "w")
		f.write(comment_text)
		f.close()

		self.svn_ops.append(	\
			DITrack.SVN.Operation (DITrack.SVN.OP_ADD, fname))

	def set_id(self, id):
		self.id = id

		self.path = os.path.join(self.db.path["data"], "i%d" % id)

		self.fname["comment"] = os.path.join(self.path, "comment")
		self.fname["info"] = os.path.join(self.path, "info")

	def set_info_header(self, name, value):
		if self.info.has_key(name):
			del self.info[name]

		self.info[name] = value

	def write(self):
		"""
		Write pending changes to disk. 
		
		
		Returns a list of SVN operations.
		"""

		# If the issue was created anew, we need to alter the database
		# before we can save it.
		if self.created:
			(id, ops) = self.db.add_issue()
			self.set_id(id)
			for o in ops:
				self.svn_ops.append(o)

		f = open(self.fname["info"], "w")
		f.write(self.info.as_string())
		f.close()

		if self.created: op = DITrack.SVN.OP_ADD
		else: op = DITrack.SVN.OP_CHANGE

		self.svn_ops.append(	\
			DITrack.SVN.Operation (op, self.fname["info"]))

		return self.svn_ops

# How was the header updated?
UPDATE_HEADER_ADD 	= 0;
UPDATE_HEADER_CHANGE 	= 1;
UPDATE_HEADER_DELETE 	= 2;

# XXX: should be named 'Update' and not tied to physical comments.
class Comment:
	def __init__(self, db, old, new, text, globals):
		self.db = db
		self.old = old
		self.new = new
		self.text = text
		self.globals = globals

	def get_changed_headers(self):
		"""
		Checks what header fields were changed and returns hash
		containing changed headers as keys, while values are 
		UPDATE_HEADER_* constants.
		"""

		changes = {}

		for k in self.old.info.keys():
			if self.new.info.has_key(k):
				if self.old.info[k] != self.new.info[k]:
					changes[k] = UPDATE_HEADER_CHANGE;
			else:
				changes[k] = UPDATE_HEADER_DELETE;

		for k in self.new.info.keys():
			if not self.old.info.has_key(k):
				changes[k] = UPDATE_HEADER_ADD;

		return changes

	def write(self):
		"Write comment data. Returns a list of svn operations."

		comment = email.Message.Message()
		comment.set_payload(self.text)

		comment["Added-on"] = \
			DITrack.Util.common.fmt_timestamp(self.globals)
		comment["Added-by"] = self.globals.username

		changed_headers = self.get_changed_headers()

		updated = (len(self.text) > 0) or changed_headers

		for k in changed_headers:

			if changed_headers[k] == UPDATE_HEADER_CHANGE:
				comment["DT-Old-" + k] = self.old.info[k]
				comment["DT-New-" + k] = self.new.info[k]

			elif changed_headers[k] == UPDATE_HEADER_DELETE:
				comment["DT-Old-" + k] = self.old.info[k]
				comment["DT-New-" + k] = ""

			elif changed_headers[k] == UPDATE_HEADER_ADD:
				comment["DT-Old-" + k] = ""
				comment["DT-New-" + k] = self.new.info[k]

			else:
				raise NotImplementedError( \
					"unknown change type: %d" % \
						changed_headers[k])

		if updated:
			self.new.add_comment(comment.as_string())
			return self.new.write()
		else:
			return []
